/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */

#include "ns3/core-module.h"
#include "ns3/helper-module.h"
#include "ns3/mobility-module.h"
#include "ns3/dce-manager-helper.h"
#include "ns3/quagga-helper.h"
#include "ns3/simulator-module.h"
#include "ns3/global-route-manager.h"
#include "ns3/node-module.h"
#include "ns3/log.h"
#include "ns3/wifi-module.h"
#include "ns3/ipv4-l3-protocol.h"
#include "ns3/ipv4-global-routing-helper.h"
#include "ns3/ipv6-l3-protocol.h"
#include "ns3/ipv6-header.h"
#include "ns3/icmpv6-header.h"
#include "ns3/ipv6-static-routing.h"
#include "ns3/udp-header.h"
#include "ns3/realtime-simulator-impl.h"

using namespace ns3;

NS_LOG_COMPONENT_DEFINE ("quagga");

void
procExit (std::string path, uint16_t pid, int retcode)
{
  std::cout << "Process exit path=" << path << " pid=" << pid << " retcode=" << retcode << std::endl;
}


static void RunIp (Ptr<Node> node, Time at, std::string str)
{
  DceApplicationHelper process;
  ApplicationContainer apps;
  process.SetBinary ("build/debug/ip");
  process.SetStackSize (1<<16);
  process.ResetArguments();
  process.ParseArguments(str.c_str ());
  apps = process.Install (node);
  apps.Start (at);
}

static void AddAddress (Ptr<Node> node, Time at, const char *name, const char *address)
{
  std::ostringstream oss;
  oss << "-f inet addr add " << address << " dev " << name;
  RunIp (node, at, oss.str ());
}

// Parameters
uint32_t nNodes = 2;
uint32_t stopTime = 60;
std::string netStack = "ns3";


int main (int argc, char *argv[])
{
  CommandLine cmd;
  cmd.AddValue ("nNodes", "Number of Router nodes", nNodes);
  cmd.AddValue ("stopTime", "Time to stop(seconds)", stopTime);
  cmd.AddValue ("netStack", "What network stack", netStack);
  cmd.Parse (argc,argv);

  // 
  //  Step 1
  //  Node Basic Configuration
  // 
  NodeContainer nodes;
  nodes.Create (nNodes);

  // 
  // 
  // Device Configuration
  // 
  // 
  MobilityHelper mobility;
  Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator> ();
  positionAlloc->Add (Vector (0.0, 0.0, 0.0));
  positionAlloc->Add (Vector (5.0, 0.0, 0.0));
  mobility.SetPositionAllocator (positionAlloc);
  mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");
  mobility.Install (nodes);

  WifiHelper wifi = WifiHelper::Default ();
  YansWifiPhyHelper phy = YansWifiPhyHelper::Default ();
  YansWifiChannelHelper phyChannel = YansWifiChannelHelper::Default ();
  NqosWifiMacHelper mac;
  phy.SetChannel (phyChannel.Create ());
  mac.SetType ("ns3::AdhocWifiMac");
  wifi.SetStandard (WIFI_PHY_STANDARD_80211a);

  NetDeviceContainer devices;
  devices = wifi.Install (phy, mac, nodes);
  phy.EnablePcapAll ("quagga");

  //  
  //  Application configuration for Nodes
  // 

  //run quagga programs in every node
  DceManagerHelper processManager;
  if (netStack == "ns3")
    {
      Ipv4AddressHelper ipv4AddrHelper;
      Ipv6AddressHelper ipv6AddrHelper;
      // Internet stack install
      InternetStackHelper stack;    // IPv4 is required for GlobalRouteMan
      stack.Install (nodes);

      ipv4AddrHelper.SetBase ("10.0.0.0", "255.255.255.0");
      Ipv4InterfaceContainer interfaces = ipv4AddrHelper.Assign (devices);
      Ipv4GlobalRoutingHelper::PopulateRoutingTables ();

      // ipv6AddrHelper.NewNetwork(Ipv6Address("2001:200:0:1::"), Ipv6Prefix(64));
      // Ipv6InterfaceContainer interfaces6 = ipv6AddrHelper.Assign (devices);

      processManager.SetNetworkStack("ns3::Ns3SocketFdFactory");
      processManager.Install (nodes);

      QuaggaHelper quagga;
      quagga.EnableOspf (nodes);
      quagga.Install (nodes);  
    }
  else if (netStack == "linux")
    {
      processManager.SetLoader ("ns3::DlmLoaderFactory");
      processManager.SetNetworkStack("ns3::LinuxSocketFdFactory",
                                     "Library", StringValue ("libnet-next-2.6.so"));
      processManager.Install (nodes);

      // IP address configuration
      AddAddress (nodes.Get (0), Seconds (0.1), "sim0", "10.0.0.1/24");
      RunIp (nodes.Get (0), Seconds (0.11), "link set lo up");
      RunIp (nodes.Get (0), Seconds (0.11), "link set sim0 up");

      AddAddress (nodes.Get (1), Seconds (0.1), "sim0", "10.0.0.2/24");
      RunIp (nodes.Get (1), Seconds (0.11), "link set lo up");
      RunIp (nodes.Get (1), Seconds (0.11), "link set sim0 up");
      RunIp (nodes.Get (0), Seconds (0.2), "link show");
      RunIp (nodes.Get (0), Seconds (0.3), "route show table all");
      RunIp (nodes.Get (0), Seconds (0.4), "addr list");

      QuaggaHelper quagga;
      quagga.Install (nodes);  

#if 1
      DceApplicationHelper process;
      ApplicationContainer apps;

      //      GenerateConfigOspf (node);
      process.ResetArguments ();
      process.SetStackSize (1<<16);
      process.SetBinary ("build/debug/ospfd");
      apps = process.Install (nodes);
      for (uint32_t i = 0; i < nodes.GetN (); i++)
        {
          apps.Get(i)->SetStartTime (Seconds (1.0 + 0.1 * i));
        }
#endif
    }

  Config::Connect ("/NodeList/*/$ns3::DceManager/Exit", MakeCallback (&procExit));

  // 
  // Step 9
  // Now It's ready to GO!
  // 
  if (stopTime != 0)
    {
      Simulator::Stop (Seconds (stopTime));
    }
  Simulator::Run ();
  Simulator::Destroy ();


  NS_LOG_INFO ("End of experiment");

  return 0;
}
