/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 Keio University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Hajime Tazaki <tazaki@sfc.wide.ad.jp>
 */
//-----------------------------------------------------------------------------
// Unit tests
//-----------------------------------------------------------------------------

#include "ns3/log.h"
#include "ns3/abort.h"
#include "ns3/attribute.h"
#include "ns3/object-factory.h"
#include "ns3/node-container.h"
#include "ns3/object-factory.h"
#include "ns3/socket-factory.h"
#include "ns3/udp-socket-factory.h"
#include "ns3/ipv4-raw-socket-factory.h"
#include "ns3/inet-socket-address.h"
#include "ns3/simulator.h"
#include "ns3/test.h"
#include "ns3/ipv4.h"
#include "ns3/simple-net-device.h"
#include "ns3/simple-channel.h"
#include "ns3/ipv4-static-routing.h"

#include "arp-l3-protocol.h"
#include "ipv4-l3-protocol.h"
#include "udp-l4-protocol.h"
#include "udp-socket-impl.h"
#include "ipv4-raw-socket-impl.h"


namespace ns3 {

class PacketInfoTagTest: public TestCase 
{
public:
  PacketInfoTagTest ();
private:
  virtual bool DoRun (void);
  void AddInternetStack (Ptr<Node> node);
  void Connect (Ptr<Node> a, Ipv4Address ipa, Ipv4Mask maskA,
		Ptr<Node> b, Ipv4Address ipb, Ipv4Mask maskB);
  void AddInterface (Ptr<Node> node, Ptr<NetDevice> dev,
		     Ipv4Address ip, Ipv4Mask mask);
  void Aggregate (Ptr<Node> node, const std::string typeId);
  void RxCb (Ptr<Socket> socket);
  void DoSendData (Ptr<Socket> socket, std::string to);
};

PacketInfoTagTest::PacketInfoTagTest ()
  : TestCase ("PacketInfoTagTest") 
{}

void
PacketInfoTagTest::AddInterface (Ptr<Node> node, Ptr<NetDevice> dev,
				 Ipv4Address ip, Ipv4Mask mask)
{
  Ptr<Ipv4> ipv4 = node->GetObject<Ipv4> ();
  uint32_t idx = ipv4->AddInterface (dev);
  Ipv4InterfaceAddress ipv4Addr = Ipv4InterfaceAddress (ip, mask);
  ipv4->AddAddress (idx, ipv4Addr);
  ipv4->SetUp (idx);
}

void
PacketInfoTagTest::Connect (Ptr<Node> a, Ipv4Address ipa, Ipv4Mask maskA,
			    Ptr<Node> b, Ipv4Address ipb, Ipv4Mask maskB)
{
  Ptr<SimpleNetDevice> devA, devB;
  { // first interface
    devA = CreateObject<SimpleNetDevice> ();
    devA->SetAddress (Mac48Address::ConvertFrom (Mac48Address::Allocate ()));
    a->AddDevice (devA);
  }
  { // second interface
    devB = CreateObject<SimpleNetDevice> ();
    devB->SetAddress (Mac48Address::ConvertFrom (Mac48Address::Allocate ()));
    b->AddDevice (devB);
  }
  Ptr<SimpleChannel> channel = CreateObject<SimpleChannel> ();
  devA->SetChannel (channel);
  devB->SetChannel (channel);

  AddInterface (a, devA, ipa, maskA);
  AddInterface (b, devB, ipb, maskB);
}

void
PacketInfoTagTest::Aggregate (Ptr<Node> node, const std::string typeId)
{
  ObjectFactory factory;
  factory.SetTypeId (typeId);
  Ptr<Object> protocol = factory.Create <Object> ();
  node->AggregateObject (protocol);
}


void
PacketInfoTagTest::AddInternetStack (Ptr<Node> node)
{
  Aggregate (node, "ns3::ArpL3Protocol");
  Aggregate (node, "ns3::Ipv4L3Protocol");
  Aggregate (node, "ns3::Icmpv4L4Protocol");
  Aggregate (node, "ns3::UdpL4Protocol");
  //Routing for Ipv4
  Ptr<Ipv4> ipv4 = node->GetObject<Ipv4> ();
  Ptr<Ipv4RoutingProtocol> ipv4Routing = CreateObject<Ipv4StaticRouting> ();
  ipv4->SetRoutingProtocol (ipv4Routing);
}

void
PacketInfoTagTest::RxCb (Ptr<Socket> socket)
{
  uint32_t availableData;
  Ptr<Packet> m_receivedPacket;

  availableData = socket->GetRxAvailable ();
  m_receivedPacket = socket->Recv (std::numeric_limits<uint32_t>::max(), 0);
  NS_ASSERT (availableData == m_receivedPacket->GetSize ());

  PacketInfoTag tag;
  bool found;
  found = m_receivedPacket->RemovePacketTag (tag);
  NS_ASSERT (found);
  tag.Print (std::cout);
}

void
PacketInfoTagTest::DoSendData (Ptr<Socket> socket, std::string to)
{
  Address realTo = InetSocketAddress (Ipv4Address (to.c_str()), 200);
  if (socket->GetInstanceTypeId () == Ipv4RawSocketImpl::GetTypeId ())
    {
      socket->SendTo (Create<Packet> (123), 0, realTo);
    }
  else if (socket->GetInstanceTypeId () == UdpSocketImpl::GetTypeId ())
    {
      NS_TEST_EXPECT_MSG_EQ (socket->SendTo (Create<Packet> (123), 0, realTo),
                             123, "XXX");
    }
}

bool
PacketInfoTagTest::DoRun (void)
{
  Ptr<Node> a = CreateObject<Node> ();
  Ptr<Node> b = CreateObject<Node> ();
  AddInternetStack (a);
  AddInternetStack (b);
  Connect (a, Ipv4Address ("10.1.1.1"), Ipv4Mask ("255.255.255.0"),
	   b, Ipv4Address ("10.1.1.1"), Ipv4Mask ("255.255.255.0"));

  // IPv4 test
  Ptr<SocketFactory> factory = a->GetObject<SocketFactory> (UdpSocketFactory::GetTypeId ());
  Ptr<Socket> socket = factory->CreateSocket ();
  InetSocketAddress local =  InetSocketAddress (Ipv4Address::GetAny (), 200);
  socket->Bind (local);
  socket->SetRecvCallback (MakeCallback (&ns3::PacketInfoTagTest::RxCb, this));

  // receive on loopback 
  Simulator::ScheduleWithContext (socket->GetNode ()->GetId (), Seconds (0),
                                  &PacketInfoTagTest::DoSendData, this, socket, "127.0.0.1");
  Simulator::Run ();

  // send from node1 and recved via csma
  Ptr<SocketFactory> factory2 = b->GetObject<SocketFactory> (UdpSocketFactory::GetTypeId ());
  Ptr<Socket> socket2 = factory2->CreateSocket ();
  Simulator::ScheduleWithContext (socket2->GetNode ()->GetId (), Seconds (0),
                                  &PacketInfoTagTest::DoSendData, this, socket, "10.1.1.1");
  Simulator::Run ();

  // ipv4 w rawsocket
  factory = a->GetObject<SocketFactory> (Ipv4RawSocketFactory::GetTypeId ());
  socket = factory->CreateSocket ();
  local =  InetSocketAddress (Ipv4Address::GetAny (), 0);
  socket->Bind (local);
  socket->SetRecvCallback (MakeCallback (&ns3::PacketInfoTagTest::RxCb, this));

  // receive on loopback 
  Simulator::ScheduleWithContext (socket->GetNode ()->GetId (), Seconds (0),
                                  &PacketInfoTagTest::DoSendData, this, socket, "127.0.0.1");
  Simulator::Run ();

  // send from node1 and recved via csma
  factory2 = b->GetObject<SocketFactory> (Ipv4RawSocketFactory::GetTypeId ());
  socket2 = factory2->CreateSocket ();
  Simulator::ScheduleWithContext (socket2->GetNode ()->GetId (), Seconds (0),
                                  &PacketInfoTagTest::DoSendData, this, socket, "10.1.1.1");
  Simulator::Run ();

  // IPv6 test
  return 0;
}

static class PacketInfoTagTestSuite : public TestSuite
{
public:
  PacketInfoTagTestSuite ();
private:
} g_packetinfotagTests;

PacketInfoTagTestSuite::PacketInfoTagTestSuite ()
  : TestSuite ("packet-info-tag", UNIT)
{
  AddTestCase (new PacketInfoTagTest ());
}

} // namespace ns3
